// Copyright (c) 2023, Tencent Inc.
// All rights reserved.
#include <stdio.h>
#include <string.h>
#include <pthread.h>
#include <unistd.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "wmpf/wmpf.h"
#include "wmpf/hardware/audio.h"
#include "wmpf/voip.h"

#include "list.h"

static pthread_mutex_t audio_player_mutex;
static pthread_mutex_t audio_record_mutex;

typedef struct {
	struct wx_audio_config config;
	const struct wx_audio_stream_out_listener *listener;
	void *user_data;
	struct wx_audio_stream_out *stream_out;
	struct listnode slist;
} AUDIO_STREAMOUT;

typedef struct {
	struct wx_audio_config config;
	const struct wx_audio_stream_in_listener *listener;
	void *user_data;
	struct wx_audio_stream_in *stream_in;
	struct listnode slist;
} AUDIO_STREAMIN;

static list_declare(audio_out_stream_list);
static list_declare(audio_in_stream_list);

extern wx_voip_session_status_t voip_status;

#define countof(x) (sizeof(x) / sizeof((x)[0]))

static const uint32_t kSupportedFormats[] = {
    WX_AUDIO_FORMAT_PCM,
};

static void audio_device_close(struct wx_device* device)
{
	printf("%s\n", __FUNCTION__);
}

static wx_error_t audio_device_getparameter(const struct wx_metadata* metadata,
						uint32_t key,
                        struct wx_metadata_entry* value_out)
{
    switch (key) {
      case WX_AUDIO_DEVICE_METADATA_MUTE:
        return WXERROR_NOT_FOUND;
      case WX_AUDIO_DEVICE_METADATA_VOLUME:
        return WXERROR_NOT_FOUND;
      case WX_AUDIO_DEVICE_METADATA_SUPPORTED_FORMATS:
        value_out->count = countof(kSupportedFormats);
        value_out->key = key;
        value_out->type = WX_METADATA_ENTRY_TYPE_U32_ARRAY,
        value_out->data.u32_array = kSupportedFormats;
        return WXERROR_OK;
      default:
        return WXERROR_NOT_FOUND;
    }
}

static wx_error_t audio_device_setparameter(struct wx_metadata* metadata,
						uint32_t key,
                        const struct wx_metadata_entry* value)
{
    switch (key) {
      case WX_AUDIO_DEVICE_METADATA_MUTE:
        return WXERROR_NOT_FOUND;
      case WX_AUDIO_DEVICE_METADATA_VOLUME:
        return WXERROR_NOT_FOUND;
      default:
        return WXERROR_NOT_FOUND;
    }
}

static wx_error_t audio_get_number_of_devices(struct wx_audio_module* module,
                                      wx_audio_device_type_t device_type,
                                      size_t* num_devices_out)
{
	printf("%s device_type = %d\n", __FUNCTION__, device_type);

	*num_devices_out = 1;

	return WXERROR_OK;
}


static wx_error_t audio_get_device_name(struct wx_audio_module* module,
                                size_t index,
                                wx_audio_device_type_t device_type,
                                char** device_name_out)
{
	printf("%s index = %d device_type =%d\n", __FUNCTION__, index, device_type);

	if (device_type == WX_AUDIO_DEVICE_IN) {
		*device_name_out = strdup("fake audio in");
	}

	if (device_type == WX_AUDIO_DEVICE_OUT) {
		*device_name_out = strdup("fake audio out");
	}

	return WXERROR_OK;
}


static void audio_stream_in_close(struct wx_audio_stream* stream)
{
	struct listnode *node;

	/*
	 * audio_in_stream_list 中的 stream 流需要关闭
	 */

	pthread_mutex_lock(&audio_record_mutex);

	list_for_each(node, &audio_in_stream_list) {

		AUDIO_STREAMIN* _stream = node_to_item(node, AUDIO_STREAMIN, slist);

		if (_stream->stream_in == (struct wx_audio_stream_in*)stream) {
	
			printf("free in stream %p\n", _stream);

			list_remove(&_stream->slist);
			free(_stream->stream_in);
			free(_stream);
			break;
		}
	}

	pthread_mutex_unlock(&audio_record_mutex);
}

static void audio_stream_out_close(struct wx_audio_stream* stream)
{
	struct listnode *node;

	/*
	 * audio_out_stream_list 中的 stream 流需要关闭
	 */

	pthread_mutex_lock(&audio_player_mutex);

	list_for_each(node, &audio_out_stream_list) {

		AUDIO_STREAMOUT* _stream = node_to_item(node, AUDIO_STREAMOUT, slist);

		if (_stream->stream_out == (struct wx_audio_stream_out*)stream) {
			
			printf("free out stream %p\n", _stream);

			list_remove(&_stream->slist);
			free(_stream->stream_out);
			free(_stream);
			break;
		}
	}

	pthread_mutex_unlock(&audio_player_mutex);
}

static wx_error_t audio_stream_in_pause(struct wx_audio_stream_in* dev)
{
	printf("%s\n", __FUNCTION__);
	return WXERROR_OK;
}

static wx_error_t audio_stream_in_resume(struct wx_audio_stream_in* dev)
{
	printf("%s\n", __FUNCTION__);
	return WXERROR_OK;
}

static wx_error_t audio_stream_out_flush(struct wx_audio_stream_out* stream)
{
	printf("%s\n", __FUNCTION__);
	return WXERROR_OK;
}

static wx_error_t audio_stream_out_pause(struct wx_audio_stream_out* dev)
{
	printf("%s\n", __FUNCTION__);
	return WXERROR_OK;
}

static wx_error_t audio_stream_out_resume(struct wx_audio_stream_out* dev)
{
	printf("%s\n", __FUNCTION__);
	return WXERROR_OK;
}

static struct wx_audio_stream_in* alloc_stream_in(void)
{
	struct wx_audio_stream_in* stream = (struct wx_audio_stream_in*)calloc(1, sizeof(struct wx_audio_stream_in));
	
	if (stream) {
		stream->common.common.size = sizeof(struct wx_audio_stream_in);
		stream->common.common.tag = WX_AUDIO_STREAM_IN_TAG;
		stream->common.common.version = 0;
		stream->common.close = audio_stream_in_close;

		stream->pause = audio_stream_in_pause;
		stream->resume = audio_stream_in_resume;
	}

	return stream;
}

static struct wx_audio_stream_out* alloc_stream_out(void)
{
	struct wx_audio_stream_out* stream = (struct wx_audio_stream_out*)calloc(1, sizeof(struct wx_audio_stream_out));
	
	if (stream) {
		stream->common.common.size = sizeof(struct wx_audio_stream_out);
		stream->common.common.tag = WX_AUDIO_STREAM_OUT_TAG;
		stream->common.common.version = 0;
		stream->common.close = audio_stream_out_close;

		stream->pause = audio_stream_out_pause;
		stream->resume = audio_stream_out_resume;
		stream->flush = audio_stream_out_flush;
	}

	return stream;
}


static void dump_audiocfg(const struct wx_audio_config* config)
{
	printf("\tsample_rate %d\n", config->sample_rate);
	printf("\tchannel_layout %x\n", config->channel_layout);
	printf("\tsample_format %x\n", config->sample_format);
	printf("\tframe_count %zu\n", config->frame_count);
	printf("\tbytes = %d\n", wx_audio_config_get_bytes_per_frame(config) );
}

static wx_error_t audio_open_input_stream(
      struct wx_audio_device_in* dev,
      const struct wx_audio_config* config,
      const struct wx_audio_stream_in_listener* listener,
      void* user_data,
      struct wx_audio_stream_in** stream_out)
{
	printf("%s user_data=%p\n", __FUNCTION__, user_data);
	dump_audiocfg(config);

	AUDIO_STREAMIN *stream = (AUDIO_STREAMIN*)calloc(1, sizeof(AUDIO_STREAMIN));

	if (stream) {
		struct wx_audio_stream_in *_stream_in = alloc_stream_in();

		if (_stream_in) {
			memcpy(&stream->config, config, sizeof(struct wx_audio_config));
			stream->listener = listener;
			stream->stream_in = _stream_in;
			stream->user_data = user_data;

			list_add_tail(&audio_in_stream_list, &stream->slist);
			*stream_out = _stream_in;

			printf("audio_open_input_stream stream = %p return %p\n", stream, _stream_in);
			return WXERROR_OK;
		}

		free(stream);		
	}
	
	return WXERROR_RESOURCE_EXHAUSTED;
}

static wx_error_t audio_open_output_stream(
      struct wx_audio_device_out* dev,
      const struct wx_audio_config* config,
      const struct wx_audio_stream_out_listener* listener,
      void* user_data,
      struct wx_audio_stream_out** stream_out)
{
	printf("%s user_data=%p\n", __FUNCTION__, user_data);
	dump_audiocfg(config);

	AUDIO_STREAMOUT *stream = (AUDIO_STREAMOUT*)calloc(1, sizeof(AUDIO_STREAMOUT));

	if (stream) {
		struct wx_audio_stream_out *_stream_out = alloc_stream_out();

		if (_stream_out) {
			memcpy(&stream->config, config, sizeof(struct wx_audio_config));
			stream->listener = listener;
			stream->stream_out = _stream_out;
			stream->user_data = user_data;

			list_add_tail(&audio_out_stream_list, &stream->slist);
			*stream_out = _stream_out;

			printf("audio_open_output_stream stream = %p return %p\n", stream, _stream_out);
			return WXERROR_OK;
		}

		free(stream);		
	}
	
	return WXERROR_RESOURCE_EXHAUSTED;
}

static struct wx_audio_device_in audio_in_device = {
	.device = {
		.common = {
			.common = {
				.tag = WX_AUDIO_DEVICE_IN_TAG,
				.size = sizeof(struct wx_audio_device_in),
				.version = 0,
			},
			.priv = NULL,
			.close = audio_device_close,
		},
		.metadata = {
			.get_parameter = audio_device_getparameter,
			.set_parameter = audio_device_setparameter,
			.priv = &audio_in_device,
		},
	},

	.open_input_stream = audio_open_input_stream,
};

static struct wx_audio_device_out audio_out_device = {
	.device = {
		.common = {
			.common = {
				.tag = WX_AUDIO_DEVICE_OUT_TAG,
				.size = sizeof(struct wx_audio_device_out),
				.version = 0,
			},
			.priv = NULL,
			.close = audio_device_close,
		},
		.metadata = {
			.get_parameter = audio_device_getparameter,
			.set_parameter = audio_device_setparameter,
			.priv = &audio_in_device,
		},
	},

	.open_output_stream = audio_open_output_stream,
};

static wx_error_t audio_open(struct wx_audio_module* module,
                     const char* id,
                     wx_audio_device_type_t device_type,
                     struct wx_audio_device** device_out)
{
	printf("%s %s device_type %d\n", __FUNCTION__, id, device_type);

	if (!strcmp(id, WX_AUDIO_DEVICE_PRIMARY)) {
		if (device_type == WX_AUDIO_DEVICE_IN) {
			*device_out = (struct wx_audio_device*)&audio_in_device;
		}

		if (device_type == WX_AUDIO_DEVICE_OUT) {
			*device_out = (struct wx_audio_device*)&audio_out_device;
		}
	}

	return WXERROR_OK;
}

struct wx_audio_module audio_module = {
	.common = {
		.common = {
			.tag = WX_AUDIO_MODULE_TAG,
			.size = sizeof(wx_audio_module_t),
			.version = 0,
		},
		.id = WX_AUDIO_MODULE_ID,
		.set_on_devices_changed = NULL,
	},

	.get_number_of_devices = audio_get_number_of_devices,
	.get_device_name = audio_get_device_name,
	.open = audio_open,
};

static int thread_exit = 0;
static pthread_t pid1 = 0, pid2 = 0;

static void hal_play_buffer(uint8_t *buffer, size_t len)
{
	/*
	 * todo: 在此实现 buffer 的实际播放
	 */

}

static void* thread_audio_player(void* data)
{
	/* 每隔 PLAYAUDIO_DELAYMS 播放一次数据，开发者可以调整此时间 */
	#define PLAYAUDIO_DELAYMS	10

	/* 每次播放数据的大小，这里根据 16000 采样 S16_LE 类型来计算 */
	#define THUNK_PLAYER_DATA_SIZE	16000/1000*PLAYAUDIO_DELAYMS*2

	int fd = open("play.pcm", O_RDWR|O_CREAT, 0700);
	if (fd < 0) {
		printf("open play.pcm fail\n");
		return NULL;
	}

	while(!thread_exit) {
		if (voip_status == WX_VOIP_SESSION_TALKING) {
			uint8_t buffer[THUNK_PLAYER_DATA_SIZE] = {0};
			struct timespec ts = {0};

			pthread_mutex_lock(&audio_player_mutex);

			/*
			 * 将链表里每一路 stream, 通过其 listener 的 data 接口来获取数据到 buffer
			 */
			struct listnode *node;
			list_for_each(node, &audio_out_stream_list) {

				AUDIO_STREAMOUT* stream;
				size_t data_len = 0;

				stream = node_to_item(node, AUDIO_STREAMOUT, slist);
				
				if (stream->config.sample_rate == WX_AUDIO_STREAM_SAMPLE_RATE_16000 && \
					stream->config.sample_format == WX_SAMPLE_FORMAT_S16) {
					/*
					 * 如果是 16000 S16 数据，直接取到 buffer 里 
					 */
					data_len = stream->listener->data(stream->stream_out, stream->user_data, ts, ts, buffer, THUNK_PLAYER_DATA_SIZE);

					if (data_len != 0) {
						write(fd, buffer, data_len);
					}
				}
				else if (stream->config.sample_rate == WX_AUDIO_STREAM_SAMPLE_RATE_44100) {
					// ...
					// 其它格式的也需要取
					//
				}
				else if (stream->config.sample_rate == WX_AUDIO_STREAM_SAMPLE_RATE_8000) {
					// ...
					// 其它格式的也需要取
					//
				}

				hal_play_buffer(buffer, data_len);
			}

			pthread_mutex_unlock(&audio_player_mutex);
		}

		usleep(1000*PLAYAUDIO_DELAYMS);
	}

	printf("exit %s\n", __FUNCTION__);
	return NULL;
}

static void* thread_audio_record(void* data)
{
	/* 每隔 RECORD_DELAYMS 给一次录音数据，开发者可以调整此时间 */
	#define RECORD_DELAYMS	10

	/* 每次给数据的大小，这里根据 16000 采样 S16_LE 类型来计算 */
	#define THUNK_RECORD_DATA_SIZE	16000/1000*RECORD_DELAYMS*2

	uint8_t buffer[THUNK_RECORD_DATA_SIZE] = {0};

	int fd = open("input.pcm", O_RDONLY, 0700);
	if (fd < 0) {
		printf("open input.pcm fail\n");
		return NULL;
	}

	while(!thread_exit) {

		if (voip_status == WX_VOIP_SESSION_TALKING) {

			struct timespec ts = {0};

			/*
			 * 这里使用pcm文件里的数据，开发者应该使用真实的录音通道数据。
			 */
			if (read(fd, buffer, THUNK_RECORD_DATA_SIZE) != THUNK_RECORD_DATA_SIZE) {
				lseek(fd, SEEK_SET, 0);

				if (read(fd, buffer, THUNK_RECORD_DATA_SIZE) != THUNK_RECORD_DATA_SIZE) {
					printf("read %d pcm bytes fail\n", THUNK_RECORD_DATA_SIZE);
					pthread_mutex_unlock(&audio_record_mutex);
					return NULL;
				}
			}

			pthread_mutex_lock(&audio_record_mutex);
			
			/*
			 * 将链表里每一路 stream_in, 使用pcm文件里的数据，通过其 listener 的 data 接口来发送
			 */
			struct listnode *node;
			list_for_each(node, &audio_in_stream_list) {

				AUDIO_STREAMIN* stream;

				stream = node_to_item(node, AUDIO_STREAMIN, slist);

				if (stream->config.sample_rate == WX_AUDIO_STREAM_SAMPLE_RATE_16000 && \
					stream->config.sample_format == WX_SAMPLE_FORMAT_S16) {
					/*
					 * PCM 文件里是 16000 S16 数据，如果 stream 是这种格式，直接发送
					 */
					stream->listener->data(stream->stream_in, stream->user_data, ts, buffer, THUNK_RECORD_DATA_SIZE);
				}
				if (stream->config.sample_rate == WX_AUDIO_STREAM_SAMPLE_RATE_8000) {
					// ...
					// PCM 文件是 16000 S16 数据，这里要 resample 再发送
					//
				}
			}
			
			pthread_mutex_unlock(&audio_record_mutex);
		}

		/*
		 * 因为是使用的文件PCM，所以使用 usleep 来进行音频时间模拟
		 */
		usleep(1000*RECORD_DELAYMS);
	}

	close(fd);
	
	printf("exit %s\n", __FUNCTION__);
	return NULL;
}

void start_audio_thread(void)
{
	/*
	 * 创建 play 线程用来播放设备收到微信端的数据
	 * 创建 record 线程用来发送数据给微信端，若开发者使用真实硬件，可能不需要创建线程，直接在录音数据的地方发送即可
	 * 
	 */
	pthread_mutex_init(&audio_player_mutex, NULL);
	pthread_mutex_init(&audio_record_mutex, NULL);

	pthread_create(&pid1, NULL, thread_audio_player, NULL);
	pthread_create(&pid2, NULL, thread_audio_record, NULL);
}

void stop_audio_thread(void)
{
	thread_exit = 1;
	pthread_join(pid1, NULL);
	pthread_join(pid2, NULL);

	pthread_mutex_destroy(&audio_player_mutex);
	pthread_mutex_destroy(&audio_record_mutex);
}