// Copyright (c) 2023, Tencent Inc.
// All rights reserved.
#pragma once

#include "wmpf/module.h"
#include "wmpf/types.h"

WX_BEGIN_DECLS

#define WX_CRYPTO_MD_CTX_TAG 0x50001

/**
 * @brief 消息摘要算法上下文
 * 该上下文的接口设计参考了 OpenSSL 的 EVP_DigestInit_ex, EVP_DigestUpdate,
 * EVP_DigestFinal_ex 函数.
 */
struct wx_crypto_md_ctx {
  struct wx_struct common;

  /**
   * @brief 传入消息片段, 并持续计算消息摘要.
   * @param input 消息片段
   * @param size 消息片段长度
   */
  wx_error_t (*update)(struct wx_crypto_md_ctx*,
                       const void* input,
                       size_t size);

  /**
   * @brief 结束消息摘要算法输入，并生成摘要.
   * @param output 存放摘要的缓冲区.
   * @param output_len 返回消息摘要长度.
   * @note 注意: 本函数没有要求传递存放摘要的缓冲区大小,
   * 这并不代表缓冲区的大小无限大.
   * output 缓冲区的大小会根据不同的消息摘要算法设置. \
   * 如对于 (HMAC-)MD5 算法, output 的大小固定 16 个字节 (MD5 是 128 位的). \
   * 如对于 (HMAC-)SHA1 算法, output 的大小固定 20 个字节 (SHA1 是 160 位的). \
   * 如对于 (HMAC-)SHA256 算法, output 的大小固定 32 个字节 (SHA1 是 256 位的).
   * \ 如对于 RSA 签名上下文, output 的大小最大 1024 个字节.
   */
  wx_error_t (*finish)(struct wx_crypto_md_ctx*, void* output, size_t* output_len);

  /**
   * @brief 销毁 wx_crypto_md_ctx 实例
   *
   */
  void (*destroy)(struct wx_crypto_md_ctx*);
};

enum wx_crypto_md_type {
  WX_CRYPTO_MD_MD5,
  WX_CRYPTO_MD_SHA1,
  WX_CRYPTO_MD_SHA256,
  WX_CRYPTO_MD_SHA512,
};

#define WX_CRYPTO_MODULE_ID "crypto"

#define WX_CRYPTO_MODULE_TAG 0x50000

typedef struct wx_crypto_module {
  struct wx_module common;

  /**
   * @brief 创建消息摘要算法上下文.
   * 消息摘要算法上下文用于生成一段消息的摘要，如 MD5, SHA1, SHA256 等算法.
   *
   * @param type 消息摘要算法类型
   */
  struct wx_crypto_md_ctx* (*md_new)(struct wx_crypto_module* module,
                                     enum wx_crypto_md_type type);

  /**
   * @brief 创建带 HMAC 的消息摘要算法上下文.
   * 消息摘要算法上下文用于生成一段消息的摘要，如 HMAC-MD5, HMAC-SHA1,
   * HMAC-SHA256 等算法.
   *
   * @param type 消息摘要算法类型
   * @param key HMAC 密钥
   * @param key_len HMAC 密钥字节数
   */
  struct wx_crypto_md_ctx* (*md_new_hmac)(struct wx_crypto_module* module,
                                          enum wx_crypto_md_type type,
                                          const void* key,
                                          size_t key_len);

  /**
   * @brief 创建用于 RSA 私钥签发证书的消息摘要算法上下文.
   * 消息摘要算法上下文用于生成一段消息的摘要，如 MD5withRSA, SHA1withRSA,
   * SHA256withRSA 等算法.
   *
   * @param type 消息摘要算法类型
   * @param private_key 私钥文本
   * @note private_key 内容格式如下：
   *   -----BEGIN RSA PRIVATE KEY-----
   *   <Base64 多行文本, 一行最多 64 个字符>
   *   -----END RSA PRIVATE KEY-----
   */
  struct wx_crypto_md_ctx* (*md_new_rsa_sign_private)(
      struct wx_crypto_module* module,
      enum wx_crypto_md_type type,
      const char* private_key);

  /**
   * @brief 创建用于 RSA 公钥签发证书的消息摘要算法上下文.
   * 消息摘要算法上下文用于生成一段消息的摘要，如 MD5withRSA, SHA1withRSA,
   * SHA256withRSA 等算法.
   *
   * @param type 消息摘要算法类型
   * @param public_key 公钥
   * @note public_key 内容格式如下：
   *   -----BEGIN PUBLIC KEY-----
   *   <Base64 多行文本, 一行最多 64 个字符>
   *   -----END PUBLIC KEY-----
   */
  struct wx_crypto_md_ctx* (*md_new_rsa_sign_public)(
      struct wx_crypto_module* module,
      enum wx_crypto_md_type type,
      const char* public_key);

} wx_crypto_module_t;

WX_END_DECLS
