// Copyright (c) 2023, Tencent Inc.
// All rights reserved.
#pragma once

#include "wmpf/hardware/device.h"
#include "wmpf/module.h"
#include "wmpf/types.h"

#include <stddef.h>
#include <stdint.h>

WX_BEGIN_DECLS

/*

音频模块包含了对音频输入输出设备、音频输入输出流的抽象。
厂商需要将:
麦克风等设备抽象为 wx_audio_device_in 对象;
扬声器等设备抽象为 wx_audio_device_out 对象.

WMPF 在获取到厂商提供的音频输入输出设备对象后, WMPF 会调用
open_input_stream 或 open_output_stream 创建音频输入输出流
来获取设备采集到的数据或者提供要播放的音频数据.

采集到的音频通过厂商调用 wx_audio_stream_in_listener::data 函数
（由 WMPF 通过 open_input_stream 传递给厂商）传递给 WMPF, 厂商需要
根据设置的音频参数，如音频格式、采样率、帧长度等设置，按时给 WMPF 传送
数据，若未按时传递数据，则会影响音频质量。

要播放的音频由厂商定时调用 wx_audio_stream_out_listener::data 函数
（由 WMPF 通过 open_output_stream 传递给厂商），要求 WMPF 提供指定
长度的数据，然后厂商拿到数据后传给设备播放。由于 WMPF 不了解设备本身对
音频数据传递间隔需求，因此厂商需要自行设置合适间隔的定时器来定时获取 WMPF
要播放的音频数据。例如厂商可以为每个 wx_audio_stream_out 创建一个线程，
每 sleep 20ms 或者其他合适的间隔后调用 WMPF 的 data 函数获取数据。

*/

#define WX_AUDIO_MODULE_ID "audio"

#define WX_AUDIO_DEVICE_PRIMARY "primary"

struct wx_audio_stream_in;
struct wx_audio_stream_out;

/**
 * @brief 音频设备参数类型
 */
typedef enum wx_audio_device_metadata_type {
  /**
   * @brief 是否静音
   * 读写
   * 类型: u32
   * 值范围: 0 表示非静音, 1 表示静音
   */
  WX_AUDIO_DEVICE_METADATA_MUTE = 0,

  /**
   * @brief 音量
   * 读写
   * 类型: u32
   * 值范围: 0~255
   */
  WX_AUDIO_DEVICE_METADATA_VOLUME = 1,

  /**
   * @brief 该音频设备所支持的音频格式
   * 只读
   * 类型: u32 array
   * 值范围: enum wx_audio_format
   */
  WX_AUDIO_DEVICE_METADATA_SUPPORTED_FORMATS = 2,
} wx_audio_device_metadata_type_t;

/**
 * @brief 音频流参数类型
 */
typedef enum wx_audio_stream_metadata_type {
  WX_AUDIO_STREAM_METADATA_VOLUME = 1,
} wx_audio_stream_metadata_type_t;

/**
 * @brief PCM 音频格式的采样值类型
 *
 */
typedef enum wx_audio_format_pcm_sub_format {
  /**
   * @brief 无符号 8 位整数 PCM
   */
  WX_AUDIO_FORMAT_PCM_SUB_U8 = 0x10,
  /**
   * @brief 小端有符号 16 位整数 PCM
   *
   */
  WX_AUDIO_FORMAT_PCM_SUB_S16LE = 0x21,
  /**
   * @brief 大端有符号 16 位整数 PCM
   *
   */
  WX_AUDIO_FORMAT_PCM_SUB_S16BE = 0x22,
  /**
   * @brief 小端有符号 24 位整数 PCM
   *
   */
  WX_AUDIO_FORMAT_PCM_SUB_S24LE = 0x31,
  /**
   * @brief 大端有符号 24 位整数 PCM
   *
   */
  WX_AUDIO_FORMAT_PCM_SUB_S24BE = 0x32,
  /**
   * @brief 小端有符号 32 位整数 PCM
   *
   */
  WX_AUDIO_FORMAT_PCM_SUB_S32LE = 0x41,
  /**
   * @brief 大端有符号 32 位整数 PCM
   *
   */
  WX_AUDIO_FORMAT_PCM_SUB_S32BE = 0x42,
  /**
   * @brief 小端有符号 32 位浮点数 PCM
   *
   */
  WX_AUDIO_FORMAT_PCM_SUB_F32LE = 0x51,
  /**
   * @brief 大端有符号 32 位浮点数 PCM
   *
   */
  WX_AUDIO_FORMAT_PCM_SUB_F32BE = 0x52,

  /**
   * @brief 小端有符号 64 位浮点数 PCM
   *
   */
  WX_AUDIO_FORMAT_PCM_SUB_F64LE = 0x61,
  /**
   * @brief 大端有符号 64 位浮点数 PCM
   *
   */
  WX_AUDIO_FORMAT_PCM_SUB_F64BE = 0x62,

  WX_AUDIO_FORMAT_PCM_SUB_MASK = 0xFF,
} wx_audio_format_pcm_sub_format_t;

/**
 * @brief 音频采样格式
 *
 * 厂商必须至少实现 WX_SAMPLE_FORMAT_S16, 用于 VoIP 通话和铃声播放.
 */
typedef enum wx_sample_format {
  WX_SAMPLE_FORMAT_UNKNOWN = 0,

  /**
   * @brief Unsigned 8-bit.
   */
  WX_SAMPLE_FORMAT_U8 = 1,

  /**
   * @brief Signed 16-bit.
   */
  WX_SAMPLE_FORMAT_S16 = 2,

  /**
   * @brief Signed 16-bit planar.
   */
  WX_SAMPLE_FORMAT_S16P = 3,

  /**
   * @brief Signed 32-bit.
   */
  WX_SAMPLE_FORMAT_S32 = 4,

  /**
   * @brief Signed 32-bit planar.
   */
  WX_SAMPLE_FORMAT_S32P = 5,

  /**
   * @brief Float 32-bit.
   */
  WX_SAMPLE_FORMAT_F32 = 6,

  /**
   * @brief Float 32-bit planar.
   */
  WX_SAMPLE_FORMAT_F32P = 7,

  /**
   * @brief Signed 24-bit planar.
   */
  WX_SAMPLE_FORMAT_S24 = 8,
} wx_sample_format_t;

WX_ALLOW_UNUSED_TYPE static inline size_t
wx_sample_format_get_bytes_per_channel(enum wx_sample_format format) {
  switch (format) {
    case WX_SAMPLE_FORMAT_U8:
      return 1;
    case WX_SAMPLE_FORMAT_S16:
    case WX_SAMPLE_FORMAT_S16P:
      return 2;
    case WX_SAMPLE_FORMAT_S24:
    case WX_SAMPLE_FORMAT_S32:
    case WX_SAMPLE_FORMAT_S32P:
    case WX_SAMPLE_FORMAT_F32:
    case WX_SAMPLE_FORMAT_F32P:
      return 4;
    default:
      return 0;
  }
}

WX_ALLOW_UNUSED_TYPE static inline bool wx_sample_format_is_planar(
    enum wx_sample_format format) {
  switch (format) {
    case WX_SAMPLE_FORMAT_S16P:
    case WX_SAMPLE_FORMAT_S32P:
    case WX_SAMPLE_FORMAT_F32P:
      return true;
    case WX_SAMPLE_FORMAT_UNKNOWN:
    case WX_SAMPLE_FORMAT_U8:
    case WX_SAMPLE_FORMAT_S16:
    case WX_SAMPLE_FORMAT_S24:
    case WX_SAMPLE_FORMAT_S32:
    case WX_SAMPLE_FORMAT_F32:
      return false;
    default:
      return false;
  }
}

WX_ALLOW_UNUSED_TYPE static inline bool wx_sample_format_is_interleaved(
    enum wx_sample_format format) {
  switch (format) {
    case WX_SAMPLE_FORMAT_U8:
    case WX_SAMPLE_FORMAT_S16:
    case WX_SAMPLE_FORMAT_S24:
    case WX_SAMPLE_FORMAT_S32:
    case WX_SAMPLE_FORMAT_F32:
      return true;
    case WX_SAMPLE_FORMAT_UNKNOWN:
    case WX_SAMPLE_FORMAT_S16P:
    case WX_SAMPLE_FORMAT_S32P:
    case WX_SAMPLE_FORMAT_F32P:
      return false;
    default:
      return false;
  }
}

WX_ALLOW_UNUSED_TYPE static inline enum wx_sample_format
wx_sample_format_to_interleaved(enum wx_sample_format format) {
  switch (format) {
    case WX_SAMPLE_FORMAT_UNKNOWN:
    case WX_SAMPLE_FORMAT_U8:
    case WX_SAMPLE_FORMAT_S16:
    case WX_SAMPLE_FORMAT_S24:
    case WX_SAMPLE_FORMAT_S32:
    case WX_SAMPLE_FORMAT_F32:
      return format;
    case WX_SAMPLE_FORMAT_S16P:
      return WX_SAMPLE_FORMAT_S16;
    case WX_SAMPLE_FORMAT_S32P:
      return WX_SAMPLE_FORMAT_S32;
    case WX_SAMPLE_FORMAT_F32P:
      return WX_SAMPLE_FORMAT_F32;
    default:
      return WX_SAMPLE_FORMAT_UNKNOWN;
  }
}

/**
 * @brief 描述 AAC 音频编码的 profile
 *
 */
typedef enum wx_audio_format_aac_sub_format {
  WX_AUDIO_FORMAT_AAC_SUB_MAIN = 0x0,
  WX_AUDIO_FORMAT_AAC_SUB_LD = 0x1,

  WX_AUDIO_FORMAT_AAC_SUB_MASK = 0xF,
} wx_audio_format_aac_sub_format_t;

/**
 * @brief 描述 G.711 音频编码的 profile
 *
 */
typedef enum wx_audio_format_g711_sub_format {
  WX_AUDIO_FORMAT_G711_SUB_NORMAL = 0x0,
  WX_AUDIO_FORMAT_G711_SUB_A_LAW = 0x1,
  WX_AUDIO_FORMAT_G711_SUB_U_LAW = 0x2,

  WX_AUDIO_FORMAT_G711_SUB_MASK = 0xF,
} wx_audio_format_g711_sub_format_t;

/**
 * @brief 描述 G.729 音频编码的 profile
 *
 */
typedef enum wx_audio_format_g729_sub_format {
  WX_AUDIO_FORMAT_G729_SUB_NORMAL = 0x0,
  WX_AUDIO_FORMAT_G729_SUB_A = 0x1,
  WX_AUDIO_FORMAT_G729_SUB_B = 0x2,

  WX_AUDIO_FORMAT_G729_SUB_MASK = 0xF,
} wx_audio_format_g729_sub_format_t;

/**
 * @brief 音频编码
 *
 * 厂商必须至少支持 WX_AUDIO_FORMAT_PCM, 用于 VoIP 通话及铃声播放.
 */
typedef enum wx_audio_format {
  WX_AUDIO_FORMAT_INVALID = 0xFFFFFFFFUL,
  WX_AUDIO_FORMAT_DEFAULT = 0,

  // clang-format off
  WX_AUDIO_FORMAT_PCM  = 0x00100000UL,
  WX_AUDIO_FORMAT_AAC  = 0x00200000UL,
  WX_AUDIO_FORMAT_MP3  = 0x00300000UL,
  WX_AUDIO_FORMAT_AC3  = 0x00400000UL,
  WX_AUDIO_FORMAT_WAV  = 0x00500000UL,
  WX_AUDIO_FORMAT_OPUS = 0x00600000UL,
  WX_AUDIO_FORMAT_SILK = 0x00700000UL,
  WX_AUDIO_FORMAT_AMR  = 0x00800000UL,
  WX_AUDIO_FORMAT_G729 = 0x00900000UL,
  WX_AUDIO_FORMAT_G711 = 0x00A00000UL,

  WX_AUDIO_FORMAT_MASK = 0xFFF00000UL,
  // clang-format on

  WX_AUDIO_FORMAT_PCM_U8 = WX_AUDIO_FORMAT_PCM | WX_AUDIO_FORMAT_PCM_SUB_U8,
  WX_AUDIO_FORMAT_PCM_S16LE =
      WX_AUDIO_FORMAT_PCM | WX_AUDIO_FORMAT_PCM_SUB_S16LE,
  WX_AUDIO_FORMAT_PCM_S16BE =
      WX_AUDIO_FORMAT_PCM | WX_AUDIO_FORMAT_PCM_SUB_S16BE,
  WX_AUDIO_FORMAT_PCM_S24LE =
      WX_AUDIO_FORMAT_PCM | WX_AUDIO_FORMAT_PCM_SUB_S24LE,
  WX_AUDIO_FORMAT_PCM_S24BE =
      WX_AUDIO_FORMAT_PCM | WX_AUDIO_FORMAT_PCM_SUB_S24BE,
  WX_AUDIO_FORMAT_PCM_S32LE =
      WX_AUDIO_FORMAT_PCM | WX_AUDIO_FORMAT_PCM_SUB_S32LE,
  WX_AUDIO_FORMAT_PCM_S32BE =
      WX_AUDIO_FORMAT_PCM | WX_AUDIO_FORMAT_PCM_SUB_S32BE,
  WX_AUDIO_FORMAT_PCM_F32LE =
      WX_AUDIO_FORMAT_PCM | WX_AUDIO_FORMAT_PCM_SUB_F32LE,
  WX_AUDIO_FORMAT_PCM_F32BE =
      WX_AUDIO_FORMAT_PCM | WX_AUDIO_FORMAT_PCM_SUB_F32BE,
  WX_AUDIO_FORMAT_PCM_F64LE =
      WX_AUDIO_FORMAT_PCM | WX_AUDIO_FORMAT_PCM_SUB_F64LE,
  WX_AUDIO_FORMAT_PCM_F64BE =
      WX_AUDIO_FORMAT_PCM | WX_AUDIO_FORMAT_PCM_SUB_F64BE,

  WX_AUDIO_FORMAT_AAC_MAIN = WX_AUDIO_FORMAT_AAC | WX_AUDIO_FORMAT_AAC_SUB_MAIN,
  WX_AUDIO_FORMAT_AAC_LD = WX_AUDIO_FORMAT_AAC | WX_AUDIO_FORMAT_AAC_SUB_LD,

  WX_AUDIO_FORMAT_G729A = WX_AUDIO_FORMAT_G729 | WX_AUDIO_FORMAT_G729_SUB_A,
  WX_AUDIO_FORMAT_G729B = WX_AUDIO_FORMAT_G729 | WX_AUDIO_FORMAT_G729_SUB_B,
} wx_audio_format_t;

/**
 * @brief 音频通道
 *
 * 描述该音频通道实际在听者的哪个方向.
 */
typedef enum wx_audio_channel {
  WX_AUDIO_CHANNEL_LEFT = 0,
  WX_AUDIO_CHANNEL_RIGHT = 1,
  WX_AUDIO_CHANNEL_CENTER = 2,
  WX_AUDIO_CHANNEL_LFE = 3,
  WX_AUDIO_CHANNEL_BACK_LEFT = 4,
  WX_AUDIO_CHANNEL_BACK_RIGHT = 5,
  WX_AUDIO_CHANNEL_LEFT_OF_CENTER = 6,
  WX_AUDIO_CHANNEL_RIGHT_OF_CENTER = 7,
  WX_AUDIO_CHANNEL_BACK_CENTER = 8,
  WX_AUDIO_CHANNEL_SIDE_LEFT = 9,
  WX_AUDIO_CHANNEL_SIDE_RIGHT = 10,

  WX_AUDIO_CHANNEL_MAX = WX_AUDIO_CHANNEL_SIDE_RIGHT,
} wx_audio_channel_t;

/**
 * @brief 音频通道布局
 *
 * 描述了当前音频包含哪些音频通道, 实际场景中布局如何. 如处理 PCM 音频流
 * 时, 就需要知道当前有几个音频通道, 并将对应通道的采样值播放到对应通道上,
 * 如选择双音频通道时, 且采样格式为 interleaved 格式，每个 audio frame
 * 将包含 2 个采样值, 第一个为左声道采样值, 第二个为右声道采样值.
 *
 * 厂商必须至少实现:
 * 1. 单音频通道 (WX_AUDIO_CHANNEL_LAYOUT_MONO), 用于 VoIP 通话
 * 2. 双音频通道 (WX_AUDIO_CHANNEL_LAYOUT_STEREO), 用于 VoIP 铃声播放.
 *
 * 其他音频通道布局为预留，厂商可不实现.
 */
typedef enum wx_audio_channel_layout {
  WX_AUDIO_CHANNEL_LAYOUT_NONE = 0,
  WX_AUDIO_CHANNEL_LAYOUT_UNSUPPORTED = 1,

  // Front C
  WX_AUDIO_CHANNEL_LAYOUT_MONO = 2,

  // Front L, Front R
  WX_AUDIO_CHANNEL_LAYOUT_STEREO = 3,

  // Front L, Front R, Back C
  WX_AUDIO_CHANNEL_LAYOUT_2_1 = 4,

  // Front L, Front R, Front C
  WX_AUDIO_CHANNEL_LAYOUT_SURROUND = 5,

  // Front L, Front R, Front C, Back C
  WX_AUDIO_CHANNEL_LAYOUT_4_0 = 6,

  // Front L, Front R, Side L, Side R
  WX_AUDIO_CHANNEL_LAYOUT_2_2 = 7,

  // Front L, Front R, Back L, Back R
  WX_AUDIO_CHANNEL_LAYOUT_QUAD = 8,

  // Front L, Front R, Front C, Side L, Side R
  WX_AUDIO_CHANNEL_LAYOUT_5_0 = 9,

  // Front L, Front R, Front C, LFE, Side L, Side R
  WX_AUDIO_CHANNEL_LAYOUT_5_1 = 10,

  // Front L, Front R, Front C, Back L, Back R
  WX_AUDIO_CHANNEL_LAYOUT_5_0_BACK = 11,

  // Front L, Front R, Front C, LFE, Back L, Back R
  WX_AUDIO_CHANNEL_LAYOUT_5_1_BACK = 12,

  // Front L, Front R, Front C, Side L, Side R, Back L, Back R
  WX_AUDIO_CHANNEL_LAYOUT_7_0 = 13,

  // Front L, Front R, Front C, LFE, Side L, Side R, Back L, Back R
  WX_AUDIO_CHANNEL_LAYOUT_7_1 = 14,

  // Front L, Front R, Front C, LFE, Side L, Side R, Front LofC, Front RofC
  WX_AUDIO_CHANNEL_LAYOUT_7_1_WIDE = 15,

  // Stereo L, Stereo R
  WX_AUDIO_CHANNEL_LAYOUT_STEREO_DOWNMIX = 16,

  // Stereo L, Stereo R, LFE
  WX_AUDIO_CHANNEL_LAYOUT_2POINT1 = 17,

  // Stereo L, Stereo R, Front C, LFE
  WX_AUDIO_CHANNEL_LAYOUT_3_1 = 18,

  // Stereo L, Stereo R, Front C, Rear C, LFE
  WX_AUDIO_CHANNEL_LAYOUT_4_1 = 19,

  // Stereo L, Stereo R, Front C, Side L, Side R, Back C
  WX_AUDIO_CHANNEL_LAYOUT_6_0 = 20,

  // Stereo L, Stereo R, Side L, Side R, Front LofC, Front RofC
  WX_AUDIO_CHANNEL_LAYOUT_6_0_FRONT = 21,

  // Stereo L, Stereo R, Front C, Rear L, Rear R, Rear C
  WX_AUDIO_CHANNEL_LAYOUT_HEXAGONAL = 22,

  // Stereo L, Stereo R, Front C, LFE, Side L, Side R, Rear Center
  WX_AUDIO_CHANNEL_LAYOUT_6_1 = 23,

  // Stereo L, Stereo R, Front C, LFE, Back L, Back R, Rear Center
  WX_AUDIO_CHANNEL_LAYOUT_6_1_BACK = 24,

  // Stereo L, Stereo R, Side L, Side R, Front LofC, Front RofC, LFE
  WX_AUDIO_CHANNEL_LAYOUT_6_1_FRONT = 25,

  // Front L, Front R, Front C, Side L, Side R, Front LofC, Front RofC
  WX_AUDIO_CHANNEL_LAYOUT_7_0_FRONT = 26,

  // Front L, Front R, Front C, LFE, Back L, Back R, Front LofC, Front RofC
  WX_AUDIO_CHANNEL_LAYOUT_7_1_WIDE_BACK = 27,

  // Front L, Front R, Front C, Side L, Side R, Rear L, Back R, Back C.
  WX_AUDIO_CHANNEL_LAYOUT_OCTAGONAL = 28,

  // Channels are not explicitly mapped to speakers.
  WX_AUDIO_CHANNEL_LAYOUT_DISCRETE = 29,

  // Front L, Front R, Front C. Front C contains the keyboard mic audio. This
  // layout is only intended for input for WebRTC. The Front C channel
  // is stripped away in the WebRTC audio input pipeline and never seen outside
  // of that.
  WX_AUDIO_CHANNEL_LAYOUT_STEREO_AND_KEYBOARD_MIC = 30,

  // Front L, Front R, Side L, Side R, LFE
  WX_AUDIO_CHANNEL_LAYOUT_4_1_QUAD_SIDE = 31,

  WX_AUDIO_CHANNEL_LAYOUT_MAX = WX_AUDIO_CHANNEL_LAYOUT_4_1_QUAD_SIDE,
} wx_audio_channel_layout_t;

WX_ALLOW_UNUSED_TYPE static inline int wx_audio_channel_layout_get_count(
    enum wx_audio_channel_layout layout) {
  switch (layout) {
    case WX_AUDIO_CHANNEL_LAYOUT_NONE:
      return 0;
    case WX_AUDIO_CHANNEL_LAYOUT_UNSUPPORTED:
      return 0;
    case WX_AUDIO_CHANNEL_LAYOUT_MONO:
      return 1;
    case WX_AUDIO_CHANNEL_LAYOUT_STEREO:
      return 2;
    case WX_AUDIO_CHANNEL_LAYOUT_2_1:
      return 3;
    case WX_AUDIO_CHANNEL_LAYOUT_SURROUND:
      return 3;
    case WX_AUDIO_CHANNEL_LAYOUT_4_0:
      return 4;
    case WX_AUDIO_CHANNEL_LAYOUT_2_2:
      return 4;
    case WX_AUDIO_CHANNEL_LAYOUT_QUAD:
      return 4;
    case WX_AUDIO_CHANNEL_LAYOUT_5_0:
      return 5;
    case WX_AUDIO_CHANNEL_LAYOUT_5_1:
      return 6;
    case WX_AUDIO_CHANNEL_LAYOUT_5_0_BACK:
      return 5;
    case WX_AUDIO_CHANNEL_LAYOUT_5_1_BACK:
      return 6;
    case WX_AUDIO_CHANNEL_LAYOUT_7_0:
      return 7;
    case WX_AUDIO_CHANNEL_LAYOUT_7_1:
      return 8;
    case WX_AUDIO_CHANNEL_LAYOUT_7_1_WIDE:
      return 8;
    case WX_AUDIO_CHANNEL_LAYOUT_STEREO_DOWNMIX:
      return 2;
    case WX_AUDIO_CHANNEL_LAYOUT_2POINT1:
      return 3;
    case WX_AUDIO_CHANNEL_LAYOUT_3_1:
      return 4;
    case WX_AUDIO_CHANNEL_LAYOUT_4_1:
      return 5;
    case WX_AUDIO_CHANNEL_LAYOUT_6_0:
      return 6;
    case WX_AUDIO_CHANNEL_LAYOUT_6_0_FRONT:
      return 6;
    case WX_AUDIO_CHANNEL_LAYOUT_HEXAGONAL:
      return 6;
    case WX_AUDIO_CHANNEL_LAYOUT_6_1:
      return 7;
    case WX_AUDIO_CHANNEL_LAYOUT_6_1_BACK:
      return 7;
    case WX_AUDIO_CHANNEL_LAYOUT_6_1_FRONT:
      return 7;
    case WX_AUDIO_CHANNEL_LAYOUT_7_0_FRONT:
      return 7;
    case WX_AUDIO_CHANNEL_LAYOUT_7_1_WIDE_BACK:
      return 8;
    case WX_AUDIO_CHANNEL_LAYOUT_OCTAGONAL:
      return 8;
    case WX_AUDIO_CHANNEL_LAYOUT_DISCRETE:
      return 0;
    case WX_AUDIO_CHANNEL_LAYOUT_STEREO_AND_KEYBOARD_MIC:
      return 3;
    case WX_AUDIO_CHANNEL_LAYOUT_4_1_QUAD_SIDE:
      return 5;
    default:
      return -1;
  }
}

WX_ALLOW_UNUSED_TYPE static inline int wx_audio_channel_layout_get_order(
    enum wx_audio_channel_layout channel_layout,
    enum wx_audio_channel channel) {
  // clang-format off
  // The channel orderings for each layout as specified by FFmpeg. Each value
  // represents the index of each channel in each layout.  Values of -1 mean the
  // channel at that index is not used for that layout. For example, the left side
  // surround sound channel in FFmpeg's 5.1 layout is in the 5th position (because
  // the order is L, R, C, LFE, LS, RS), so
  // CHANNEL_ORDINGS[WX_AUDIO_CHANNEL_LAYOUT_5_1][WX_AUDIO_CHANNEL_SIDE_LEFT] = 4;
  static const int CHANNEL_ORDINGS[WX_AUDIO_CHANNEL_LAYOUT_MAX + 1][WX_AUDIO_CHANNEL_MAX + 1] = {
    // FL | FR | FC | LFE | BL | BR | FLofC | FRofC | BC | SL | SR

    // WX_AUDIO_CHANNEL_LAYOUT_NONE
    {  -1 , -1 , -1 , -1  , -1 , -1 , -1    , -1    , -1 , -1 , -1 },

    // WX_AUDIO_CHANNEL_LAYOUT_UNSUPPORTED
    {  -1 , -1 , -1 , -1  , -1 , -1 , -1    , -1    , -1 , -1 , -1 },

    // WX_AUDIO_CHANNEL_LAYOUT_MONO
    {  -1 , -1 , 0  , -1  , -1 , -1 , -1    , -1    , -1 , -1 , -1 },

    // WX_AUDIO_CHANNEL_LAYOUT_STEREO
    {  0  , 1  , -1 , -1  , -1 , -1 , -1    , -1    , -1 , -1 , -1 },

    // WX_AUDIO_CHANNEL_LAYOUT_2_1
    {  0  , 1  , -1 , -1  , -1 , -1 , -1    , -1    , 2  , -1 , -1 },

    // WX_AUDIO_CHANNEL_LAYOUT_SURROUND
    {  0  , 1  , 2  , -1  , -1 , -1 , -1    , -1    , -1 , -1 , -1 },

    // WX_AUDIO_CHANNEL_LAYOUT_4_0
    {  0  , 1  , 2  , -1  , -1 , -1 , -1    , -1    , 3  , -1 , -1 },

    // WX_AUDIO_CHANNEL_LAYOUT_2_2
    {  0  , 1  , -1 , -1  , -1 , -1 , -1    , -1    , -1 , 2  ,  3 },

    // WX_AUDIO_CHANNEL_LAYOUT_QUAD
    {  0  , 1  , -1 , -1  , 2  , 3  , -1    , -1    , -1 , -1 , -1 },

    // WX_AUDIO_CHANNEL_LAYOUT_5_0
    {  0  , 1  , 2  , -1  , -1 , -1 , -1    , -1    , -1 , 3  ,  4 },

    // WX_AUDIO_CHANNEL_LAYOUT_5_1
    {  0  , 1  , 2  , 3   , -1 , -1 , -1    , -1    , -1 , 4  ,  5 },

    // FL | FR | FC | LFE | BL | BR | FLofC | FRofC | BC | SL | SR

    // WX_AUDIO_CHANNEL_LAYOUT_5_0_BACK
    {  0  , 1  , 2  , -1  , 3  , 4  , -1    , -1    , -1 , -1 , -1 },

    // WX_AUDIO_CHANNEL_LAYOUT_5_1_BACK
    {  0  , 1  , 2  , 3   , 4  , 5  , -1    , -1    , -1 , -1 , -1 },

    // WX_AUDIO_CHANNEL_LAYOUT_7_0
    {  0  , 1  , 2  , -1  , 5  , 6  , -1    , -1    , -1 , 3  ,  4 },

    // WX_AUDIO_CHANNEL_LAYOUT_7_1
    {  0  , 1  , 2  , 3   , 6  , 7  , -1    , -1    , -1 , 4  ,  5 },

    // WX_AUDIO_CHANNEL_LAYOUT_7_1_WIDE
    {  0  , 1  , 2  , 3   , -1 , -1 , 6     , 7     , -1 , 4  ,  5 },

    // WX_AUDIO_CHANNEL_LAYOUT_STEREO_DOWNMIX
    {  0  , 1  , -1 , -1  , -1 , -1 , -1    , -1    , -1 , -1 , -1 },

    // WX_AUDIO_CHANNEL_LAYOUT_2POINT1
    {  0  , 1  , -1 ,  2  , -1 , -1 , -1    , -1    , -1 , -1 , -1 },

    // WX_AUDIO_CHANNEL_LAYOUT_3_1
    {  0  , 1  ,  2 ,  3  , -1 , -1 , -1    , -1    , -1 , -1 , -1 },

    // WX_AUDIO_CHANNEL_LAYOUT_4_1
    {  0  , 1  ,  2 ,  4  , -1 , -1 , -1    , -1    ,  3 , -1 , -1 },

    // WX_AUDIO_CHANNEL_LAYOUT_6_0
    {  0  , 1  , 2  , -1  , -1 , -1 , -1    , -1    ,  5 , 3  ,  4 },

    // WX_AUDIO_CHANNEL_LAYOUT_6_0_FRONT
    {  0  , 1  , -1 , -1  , -1 , -1 ,  4    ,  5    , -1 , 2  ,  3 },

    // FL | FR | FC | LFE | BL | BR | FLofC | FRofC | BC | SL | SR

    // WX_AUDIO_CHANNEL_LAYOUT_HEXAGONAL
    {  0  , 1  , 2  , -1  , 3  , 4  , -1    , -1    ,  5 , -1 , -1 },

    // WX_AUDIO_CHANNEL_LAYOUT_6_1
    {  0  , 1  , 2  , 3   , -1 , -1 , -1    , -1    ,  6 , 4  ,  5 },

    // WX_AUDIO_CHANNEL_LAYOUT_6_1_BACK
    {  0  , 1  , 2  , 3   , 4  , 5  , -1    , -1    ,  6 , -1 , -1 },

    // WX_AUDIO_CHANNEL_LAYOUT_6_1_FRONT
    {  0  , 1  , -1 , 6   , -1 , -1 , 4     , 5     , -1 , 2  ,  3 },

    // WX_AUDIO_CHANNEL_LAYOUT_7_0_FRONT
    {  0  , 1  , 2  , -1  , -1 , -1 , 5     , 6     , -1 , 3  ,  4 },

    // WX_AUDIO_CHANNEL_LAYOUT_7_1_WIDE_BACK
    {  0  , 1  , 2  , 3   , 4  , 5  , 6     , 7     , -1 , -1 , -1 },

    // WX_AUDIO_CHANNEL_LAYOUT_OCTAGONAL
    {  0  , 1  , 2  , -1  , 5  , 6  , -1    , -1    ,  7 , 3  ,  4 },

    // WX_AUDIO_CHANNEL_LAYOUT_DISCRETE
    {  -1 , -1 , -1 , -1  , -1 , -1 , -1    , -1    , -1 , -1 , -1 },

    // WX_AUDIO_CHANNEL_LAYOUT_STEREO_AND_KEYBOARD_MIC
    {  0  , 1  , 2  , -1  , -1 , -1 , -1    , -1    , -1 , -1 , -1 },

    // WX_AUDIO_CHANNEL_LAYOUT_4_1_QUAD_SIDE
    {  0  , 1  , -1 ,  4  , -1 , -1 , -1    , -1    , -1 , 2  ,  3 },

    // FL | FR | FC | LFE | BL | BR | FLofC | FRofC | BC | SL | SR
  };
  // clang-format on

  if (0 <= channel_layout && channel_layout <= WX_AUDIO_CHANNEL_LAYOUT_MAX &&
      0 <= channel && channel <= WX_AUDIO_CHANNEL_MAX) {
    return CHANNEL_ORDINGS[channel_layout][channel];
  } else {
    return -1;
  }
}

/**
 * @brief 用于描述音频采样率，仅PCM格式有效
 *
 */
typedef enum wx_audio_stream_sample_rate {
  WX_AUDIO_STREAM_SAMPLE_RATE_8000 = 8000,   /* 8K samplerate */
  WX_AUDIO_STREAM_SAMPLE_RATE_12000 = 12000, /* 12K samplerate */
  WX_AUDIO_STREAM_SAMPLE_RATE_11025 = 11025, /* 11.025K samplerate */
  WX_AUDIO_STREAM_SAMPLE_RATE_16000 = 16000, /* 16K samplerate */
  WX_AUDIO_STREAM_SAMPLE_RATE_22050 = 22050, /* 22.050K samplerate */
  WX_AUDIO_STREAM_SAMPLE_RATE_24000 = 24000, /* 24K samplerate */
  WX_AUDIO_STREAM_SAMPLE_RATE_32000 = 32000, /* 32K samplerate */
  WX_AUDIO_STREAM_SAMPLE_RATE_44100 = 44100, /* 44.1K samplerate */
  WX_AUDIO_STREAM_SAMPLE_RATE_48000 = 48000, /* 48K samplerate */
  WX_AUDIO_STREAM_SAMPLE_RATE_64000 = 64000, /* 64K samplerate */
  WX_AUDIO_STREAM_SAMPLE_RATE_96000 = 96000, /* 96K samplerate */
} wx_audio_stream_sample_rate_t;

typedef enum wx_audio_stream_type {
  WX_AUDIO_STREAM_DEFAULT = -1,
  WX_AUDIO_STREAM_VOICE_CALL = 0,
  WX_AUDIO_STREAM_SYSTEM = 1,
  WX_AUDIO_STREAM_RING = 2,
  WX_AUDIO_STREAM_MUSIC = 3,
  WX_AUDIO_STREAM_ALARM = 4,
  WX_AUDIO_STREAM_NOTIFICATION = 5,
} wx_audio_stream_type_t;

/**
 * @brief 音频流配置参数
 *
 */
typedef struct wx_audio_config {
  struct wx_struct common;

  /**
   * @brief 音频流采样率
   *
   * 音频输入流和音频输出流均必须至少支持 8000、16000、44100Hz 的采样率.
   * 如果你的音频设备实际上不支持某一种采样率，你需要在实现中自行重采样.
   */
  uint32_t sample_rate;

  /**
   * @brief 音频流通道布局
   *
   * 音频输入流和音频输出流均必须至少支持 WX_AUDIO_CHANNEL_LAYOUT_MONO 和
   * WX_AUDIO_CHANNEL_LAYOUT_STERERO.
   */
  wx_audio_channel_layout_t channel_layout;

  /**
   * @brief 音频流数据格式
   *
   * 音频输入流和音频输出流均必须至少支持 WX_AUDIO_FORMAT_PCM.
   */
  wx_audio_format_t format;

  /**
   * @brief 音频流采样格式
   *
   * 音频流为 PCM 格式时，采样格式必须要传递实际采样格式，不可以传递
   * WX_SAMPLE_FORMAT_UNKNOWN.
   * 音频流不为 PCM 格式时，该项可填 WX_SAMPLE_FORMAT_UNKNOWN.
   *
   * 音频输入流和音频输出流均必须至少支持 WX_SAMPLE_FORMAT_S16.
   */
  wx_sample_format_t sample_format;
  size_t frame_count;
} wx_audio_config_t;

WX_ALLOW_UNUSED_TYPE static inline size_t wx_audio_config_get_bytes_per_frame(
    const struct wx_audio_config* config) {
  return wx_audio_channel_layout_get_count(config->channel_layout) *
         wx_sample_format_get_bytes_per_channel(config->sample_format);
}

WX_ALLOW_UNUSED_TYPE static inline size_t wx_audio_config_get_bytes_per_buffer(
    const struct wx_audio_config* config) {
  return wx_audio_config_get_bytes_per_frame(config) * config->frame_count;
}

/**
 * @brief 通用音频流抽象类
 *
 */
typedef struct wx_audio_stream {
  struct wx_struct common;

  struct wx_metadata metadata;

  void (*close)(struct wx_audio_stream* stream);
} wx_audio_stream_t;

#define WX_AUDIO_STREAM_IN_LISTENER_TAG 0x10001

typedef struct wx_audio_stream_in_listener {
  struct wx_struct common;

  /**
   * @brief 采集到的音频数据回调
   * @param user_data 用户自定义参数，和调用 struct
   * @param capture_time 采集到的音频的时刻
   * @param buffer 音频数据，数据格式可以通过 metadata 获取
   * @param bytes 音频数据长度
   */
  void (*data)(struct wx_audio_stream_in* stream,
               void* user_data,
               struct timespec capture_time,
               const void* buffer,
               size_t buffer_size);

  /**
   * @brief 通知 wx_audio_stream_in 在录制时出现错误
   * @param stream 出现错误的音频输入流
   * @param user_data 通过 open_input_stream 传入的 user_data
   */
  void (*error)(struct wx_audio_stream_in* stream,
                void* user_data,
                wx_error_t status);
} wx_audio_stream_in_listener_t;

#define WX_AUDIO_STREAM_IN_TAG 0x10002

/**
 * @brief 音频输入流（采集音频）抽象类
 *
 */
typedef struct wx_audio_stream_in {
  /**
   * @brief 音频输入流抽象类继承自通用音频流抽象类
   * 该变量必须是 struct wx_audio_stream_in 的第一个成员，因为用户需要可以将
   * struct wx_audio_stream_in 类转换为 struct wx_audio_stream 类.
   */
  struct wx_audio_stream common;

  /**
   * @brief 暂停音频录制
   * 暂停音频录制后, wx_audio_stream_in_listener::data 将暂停调用
   * @return 操作结果
   *   - WXERROR_FAILED_PRECONDITION: 当前不是播放状态
   */
  wx_error_t (*pause)(struct wx_audio_stream_in* dev);

  /**
   * @brief 恢复音频录制
   * @return 操作结果
   *   - WXERROR_FAILED_PRECONDITION: 当前不是暂停状态
   */
  wx_error_t (*resume)(struct wx_audio_stream_in* dev);

} wx_audio_stream_in_t;

#define WX_AUDIO_STREAM_OUT_LISTENER_TAG 0x10003

typedef struct wx_audio_stream_out_listener {
  struct wx_struct common;

  /**
   * @brief 向 wx_audio_stream_out 提供所要播放的音频数据
   * @param stream 要播放音频的音频输出流
   * @param user_data 通过 open_output_stream 传入的 user_data
   * @param delay 当前写入的音频在 delay 时间后播放; 即上一次调用 data
   * 时写入的还未播放的音频的时长 (默认可以传递 0).
   * @param delay_timestamp 测量 delay 的时刻 (默认传递当前时刻)
   * @param buffer 需要写入要播放的数据的缓冲区
   * @param buffer_size 缓冲区大小，最多写入的数据字节数
   * @return 写入的字节数, 0 表示要播放的音频数据已经结束, 不需要再调用 data
   * 函数.
   */
  size_t (*data)(struct wx_audio_stream_out* stream,
                 void* user_data,
                 struct timespec delay,
                 struct timespec delay_timestamp,
                 void* buffer,
                 size_t buffer_size);

  /**
   * @brief 通知 wx_audio_stream_out 在播放时出现错误
   * @param stream 出现错误的音频输出流
   * @param user_data 通过 open_output_stream 传入的 user_data
   */
  void (*error)(struct wx_audio_stream_out* stream,
                void* user_data,
                wx_error_t status);
} wx_audio_stream_out_listener_t;

#define WX_AUDIO_STREAM_OUT_TAG 0x10004

/**
 * @brief 音频输出流（播放音频）抽象类
 *
 */
typedef struct wx_audio_stream_out {
  /**
   * @brief 音频输出流抽象类继承自通用音频流抽象类
   * 该变量必须是 struct wx_audio_stream_out 的第一个成员，因为用户需要可以将
   * struct wx_audio_stream_out 类转换为 struct wx_audio_stream 类.
   */
  struct wx_audio_stream common;

  /**
   * @brief 暂停音频播放
   * 暂停音频播放后，已经写入的音频数据需要被保留，以便 resume 调用后恢复播放。
   * @return 操作结果
   *   - WXERROR_FAILED_PRECONDITION: 当前不是播放状态
   */
  wx_error_t (*pause)(struct wx_audio_stream_out* dev);

  /**
   * @brief 恢复音频播放
   * @return 操作结果
   *   - WXERROR_FAILED_PRECONDITION: 当前不是暂停状态
   */
  wx_error_t (*resume)(struct wx_audio_stream_out* dev);

  /**
   * @brief 清空音频流缓存
   * @return 操作是否成功
   */
  wx_error_t (*flush)(struct wx_audio_stream_out* stream);

} wx_audio_stream_out_t;

/**
 * @brief 音频设备抽象类
 *
 */
typedef struct wx_audio_device {
  /**
   * @brief 音频设备抽象类继承自设备抽象类
   * 该变量必须是 struct wx_audio_device 的第一个成员，因为用户需要可以将
   * struct wx_audio_device 类转换为 struct wx_device 类.
   */
  struct wx_device common;

  /**
   * @brief 音频设备的属性集
   * key 列表参考 wx_audio_device_metadata_type.
   */
  struct wx_metadata metadata;

} wx_audio_device_t;

/**
 * @brief 音频设备类型
 *
 */
typedef enum wx_audio_device_type {
  /**
   * @brief 音频输入设备
   *
   */
  WX_AUDIO_DEVICE_IN = 0,

  /**
   * @brief 音频输出设备
   *
   */
  WX_AUDIO_DEVICE_OUT = 1,
} wx_audio_device_type_t;

#define WX_AUDIO_DEVICE_IN_TAG 0x10005

/**
 * @brief 音频输入设备
 *
 * 厂商必须至少提供一个音频输入设备,
 * 并允许该音频输入设备同时支持至少一个音频输入流.
 *
 * 厂商可以通过 WX_AUDIO_DEVICE_METADATA_SUPPORTED_FORMAT
 * 指定支持采集的音频流数据格式， 但是 sample_format 应当支持任意格式;
 * 音频采样率 sample_rate 也应支持任意频率.
 * 如果设备实际上不支持指定的采样格式和采样率, 厂商需要自行实现音频的重采样.
 * 采样格式目前主要使用 WX_SAMPLE_FORMAT_S16; 音频采样率主要使用 8000Hz,
 * 16000Hz, 44100Hz.
 */
typedef struct wx_audio_device_in {
  struct wx_audio_device device;

  /**
   * @brief 创建音频输入流，并同时开始音频的录制
   *
   * @param config 音频录制配置
   * @param listener 音频输入流回调, 需要保证生命周期长于返回的音频流
   * @param user_data 音频输入流回调用户数据, 需要保证生命周期长于返回的音频流
   * @param stream_out 返回音频输入流
   * @return 操作结果
   *   - WXERROR_INVALID_ARGUMENT: 参数不合法
   *   - WXERROR_NOT_FOUND: 音频格式 config->format 不受支持
   *   - WXERROR_INTERNAL: 内部错误
   */
  wx_error_t (*open_input_stream)(
      struct wx_audio_device_in* dev,
      const struct wx_audio_config* config,
      const struct wx_audio_stream_in_listener* listener,
      void* user_data,
      struct wx_audio_stream_in** stream_out);

} wx_audio_device_in_t;

#define WX_AUDIO_DEVICE_OUT_TAG 0x10006

/**
 * @brief 音频输出设备
 *
 * 厂商必须至少提供一个音频输出设备,
 * 并允许该音频输出设备能同时支持多个音频输出流,
 * 并自行处理不同音频输出流的不同音频格式解码、混合.
 *
 * 厂商可以通过 WX_AUDIO_DEVICE_METADATA_SUPPORTED_FORMAT
 * 指定支持播放的音频流数据格式， 但是 sample_format 应当支持任意格式;
 * 音频采样率 sample_rate 也应支持任意频率.
 * 如果设备实际上不支持指定的采样格式和采样率, 厂商需要自行实现音频的重采样.
 * 采样格式目前主要使用 WX_SAMPLE_FORMAT_S16; 音频采样率主要使用 8000Hz,
 * 16000Hz, 44100Hz.
 */
typedef struct wx_audio_device_out {
  struct wx_audio_device device;

  /**
   * @brief 创建在此音频输出设备上播放音频的音频输出流
   * @param config 要播放的音频的参数
   * @param listener 音频播放数据源
   * @param user_data 传递给 listener
   * 的自定义数据，需要保证该指针指向的内容生命周期比 wx_audio_stream_out 长
   * @param stream_out 返回音频输出流
   * @return 操作结果
   *   - WXERROR_INVALID_ARGUMENT: 参数不合法
   *   - WXERROR_NOT_FOUND: 音频格式 config->format 不受支持
   *   - WXERROR_INTERNAL: 内部错误
   */
  wx_error_t (*open_output_stream)(
      struct wx_audio_device_out* dev,
      const struct wx_audio_config* config,
      const struct wx_audio_stream_out_listener* listener,
      void* user_data,
      struct wx_audio_stream_out** stream_out);

} wx_audio_device_out_t;

#define WX_AUDIO_MODULE_TAG 0x10000

/**
 * @brief 音频模块抽象类
 *
 */
typedef struct wx_audio_module {
  struct wx_module common;

  /**
   * @brief 获取音频设备数量
   * @param device_type 音频设备类型
   * @param num_devices_out 保存当前时刻指定音频设备类型的音频设备数量
   * @return 获取结果
   *   - WXERROR_INVALID_ARGUMENT: num_devices_out 为空
   */
  wx_error_t (*get_number_of_devices)(struct wx_audio_module* module,
                                      wx_audio_device_type_t device_type,
                                      size_t* num_devices_out);

  /**
   * @brief 获取音频设备名称
   * @param index 指定音频设备类型的音频设备编号，范围为 [0,
   * get_num_audio_devices(device_type))
   * @param device_type 音频设备类型
   * @return 获取结果
   *   - WXERROR_INVALID_ARGUMENT: device_name_out 为空
   *   - WXERROR_OUT_OF_RANGE: index 超出范围
   */
  wx_error_t (*get_device_name)(struct wx_audio_module* module,
                                size_t index,
                                wx_audio_device_type_t device_type,
                                char** device_name_out);

  /**
   * @brief 获取音频设备信息
   *
   * 该函数用于返回一些在不用打开音频设备的情况下就能获取到的音频设备信息,
   * 如支持的音频编码格式.
   *
   * @param id 音频设备名称, 若传入 WX_AUDIO_DEVICE_PRIMARY,
   * 则必须返回一个可用的音频设备.
   * @param device_type 音频设备类型.
   */
  wx_error_t (*get_device_info)(struct wx_audio_module* module,
                                const char* id,
                                wx_audio_device_type_t device_type,
                                const struct wx_metadata** metadata_out);

  /**
   * @brief 打开音频设备
   *
   * 厂商必须至少为音频输入类型和音频输出类型各提供一个音频设备，且该设备名称必须为
   * WX_AUDIO_DEVICE_PRIMARY
   *
   * @param id 音频设备名称, 若传入 WX_AUDIO_DEVICE_PRIMARY,
   * 则必须返回一个可用的音频设备.
   * @param device_type 音频设备类型.
   * @param device_out 音频设备对象, 可以是 wx_audio_device_in 对象, 也可以是
   * wx_audio_device_out 对象.
   */
  wx_error_t (*open)(struct wx_audio_module* module,
                     const char* id,
                     wx_audio_device_type_t device_type,
                     struct wx_audio_device** device_out);

} wx_audio_module_t;

WX_END_DECLS
