// Copyright (c) 2023, Tencent Inc.
// All rights reserved.
#pragma once

#include <stddef.h>
#include <stdint.h>

#include "wmpf/error.h"
#include "wmpf/macros.h"
#include "wmpf/types.h"

WX_BEGIN_DECLS

typedef enum wx_metadata_entry_type {
  /**
   * @brief 无符号 32 位整型
   */
  WX_METADATA_ENTRY_TYPE_U32 = 0,
  /**
   * @brief 有符号 32 位整型
   */
  WX_METADATA_ENTRY_TYPE_I32 = 1,
  /**
   * @brief 单精度浮点型
   */
  WX_METADATA_ENTRY_TYPE_F32 = 2,
  /**
   * @brief 双精度浮点型
   */
  WX_METADATA_ENTRY_TYPE_F64 = 3,
  /**
   * @brief 字符串
   */
  WX_METADATA_ENTRY_TYPE_STRING = 4,
  /**
   * @brief 无符号 8 位整型数组
   */
  WX_METADATA_ENTRY_TYPE_U8_ARRAY = 5,
  /**
   * @brief 有符号 32 位整型数组
   */
  WX_METADATA_ENTRY_TYPE_I32_ARRAY = 6,
  /**
   * @brief 无符号 32 位整型数组
   */
  WX_METADATA_ENTRY_TYPE_U32_ARRAY = 7,
  /**
   * @brief 单精度浮点型数组
   */
  WX_METADATA_ENTRY_TYPE_F32_ARRAY = 8,
  /**
   * @brief 有符号 64 位整型数组
   */
  WX_METADATA_ENTRY_TYPE_I64_ARRAY = 9,
  /**
   * @brief 双精度浮点型数组
   */
  WX_METADATA_ENTRY_TYPE_F64_ARRAY = 10,
} wx_metadata_entry_type_t;

typedef struct wx_metadata_entry {
  /**
   * @brief 元数据种类，由使用方定义.
   */
  uint32_t key;

  /**
   * @brief 元数据值类型
   * 与元数据种类对应的类型应当一致.
   */
  enum wx_metadata_entry_type type;

  /**
   * @brief 元数据值元素个数
   * 仅当元数据值为数组时有效，否则应当为 0.
   */
  size_t count;

  union {
    uint32_t u32;
    int32_t i32;
    float f32;
    double f64;
    const char* string;
    const uint8_t* u8_array;
    const int32_t* i32_array;
    const uint32_t* u32_array;
    const float* f32_array;
    const int64_t* i64_array;
    const double* f64_array;
  } data;
} wx_metadata_entry_t;

WX_ALLOW_UNUSED_TYPE static inline void wx_metadata_entry_set_u32(
    struct wx_metadata_entry* entry,
    uint32_t key,
    uint32_t value) {
  entry->key = key;
  entry->type = WX_METADATA_ENTRY_TYPE_U32;
  entry->count = 0;
  entry->data.u32 = value;
}

WX_ALLOW_UNUSED_TYPE static inline void wx_metadata_entry_set_i32(
    struct wx_metadata_entry* entry,
    uint32_t key,
    int32_t value) {
  entry->key = key;
  entry->type = WX_METADATA_ENTRY_TYPE_I32;
  entry->count = 0;
  entry->data.i32 = value;
}

WX_ALLOW_UNUSED_TYPE static inline void wx_metadata_entry_set_f32(
    struct wx_metadata_entry* entry,
    uint32_t key,
    float value) {
  entry->key = key;
  entry->type = WX_METADATA_ENTRY_TYPE_F32;
  entry->count = 0;
  entry->data.f32 = value;
}

WX_ALLOW_UNUSED_TYPE static inline void wx_metadata_entry_set_f64(
    struct wx_metadata_entry* entry,
    uint32_t key,
    double value) {
  entry->key = key;
  entry->type = WX_METADATA_ENTRY_TYPE_F64;
  entry->count = 0;
  entry->data.f64 = value;
}

WX_ALLOW_UNUSED_TYPE static inline void wx_metadata_entry_set_string(
    struct wx_metadata_entry* entry,
    uint32_t key,
    const char* value) {
  entry->key = key;
  entry->type = WX_METADATA_ENTRY_TYPE_STRING;
  entry->count = 0;
  entry->data.string = value;
}

WX_ALLOW_UNUSED_TYPE static inline void wx_metadata_entry_set_u8_array(
    struct wx_metadata_entry* entry,
    uint32_t key,
    const uint8_t* array,
    size_t count) {
  entry->key = key;
  entry->type = WX_METADATA_ENTRY_TYPE_U8_ARRAY;
  entry->count = count;
  entry->data.u8_array = array;
}

WX_ALLOW_UNUSED_TYPE static inline void wx_metadata_entry_set_i32_array(
    struct wx_metadata_entry* entry,
    uint32_t key,
    const int32_t* array,
    size_t count) {
  entry->key = key;
  entry->type = WX_METADATA_ENTRY_TYPE_I32_ARRAY;
  entry->count = count;
  entry->data.i32_array = array;
}

WX_ALLOW_UNUSED_TYPE static inline void wx_metadata_entry_set_u32_array(
    struct wx_metadata_entry* entry,
    uint32_t key,
    const uint32_t* array,
    size_t count) {
  entry->key = key;
  entry->type = WX_METADATA_ENTRY_TYPE_U32_ARRAY;
  entry->count = count;
  entry->data.u32_array = array;
}

WX_ALLOW_UNUSED_TYPE static inline void wx_metadata_entry_set_f32_array(
    struct wx_metadata_entry* entry,
    uint32_t key,
    const float* array,
    size_t count) {
  entry->key = key;
  entry->type = WX_METADATA_ENTRY_TYPE_F32_ARRAY;
  entry->count = count;
  entry->data.f32_array = array;
}

WX_ALLOW_UNUSED_TYPE static inline void wx_metadata_entry_set_i64_array(
    struct wx_metadata_entry* entry,
    uint32_t key,
    const int64_t* array,
    size_t count) {
  entry->key = key;
  entry->type = WX_METADATA_ENTRY_TYPE_I64_ARRAY;
  entry->count = count;
  entry->data.i64_array = array;
}

WX_ALLOW_UNUSED_TYPE static inline void wx_metadata_entry_set_f64_array(
    struct wx_metadata_entry* entry,
    uint32_t key,
    const double* array,
    size_t count) {
  entry->key = key;
  entry->type = WX_METADATA_ENTRY_TYPE_F64_ARRAY;
  entry->count = count;
  entry->data.f64_array = array;
}

/**
 * @brief 描述一个对象的属性
 *
 * 如果一个对象有多个属性可供访问或修改, 则会实现该接口.
 */
typedef struct wx_metadata {
  /**
   * @brief 访问属性
   * @param key 属性名称
   * @param value_out 保存属性值
   * @return 获取是否成功
   *   - WXERROR_INVALID_ARGUMENT: metadata 或 value_out 为 NULL
   *   - WXERROR_NOT_FOUND: key 不存在
   */
  wx_error_t (*get_parameter)(const struct wx_metadata* metadata,
                              uint32_t key,
                              struct wx_metadata_entry* value_out);

  /**
   * @brief 修改属性
   * @param key 属性名称
   * @param value 要设置的属性值
   * @return 设置是否成功
   *   - WXERROR_INVALID_ARGUMENT: 属性类型不匹配
   *   - WXERROR_NOT_FOUND: key 不存在
   *   - WXERROR_FAILED_PRECONDITION: 当前不能设置给定属性
   *   - WXERROR_OUT_OF_RANGE: 属性值不合法
   */
  wx_error_t (*set_parameter)(struct wx_metadata* metadata,
                              uint32_t key,
                              const struct wx_metadata_entry* value);

  /**
   * @brief WMPF 的自定义内部信息
   * 该参数提供给 WMPF 实现层访问 wx_metadata 所对应要操作的对象或信息.
   * 厂商不应当访问或修改该变量.
   */
  void* priv;
} wx_metadata_t;

typedef struct wx_module {
  struct wx_struct common;

  /**
   * @brief 模块 ID
   * 应当为 WX_<...>_MODULE_ID
   */
  const char* id;

  /**
   * @brief 设置设备列表变化回调
   */
  wx_error_t (*set_on_devices_changed)(struct wx_module* module);

} wx_module_t;

/**
 * @brief 根据 id 获取厂商接口模块实例
 *
 * 厂商需要实现该函数，并提供以下模块:
 *  - 音频模块 (ID=WX_AUDIO_MODULE_ID, wmpf/hardware/audio.h)
 *  - 摄像头模块 (ID=WX_CAMERA_MODULE_ID, wmpf/hardware/camera.h)
 *  - 显示模块 (ID=WX_DISPLAY_MODULE_ID, wmpf/hardware/display.h)
 *  - 视频模块 (ID=WX_VIDEO_MODULE_ID, wmpf/hardware/video.h)
 * 以上模块需要由厂商实现，并被 WMPF 通过这些接口访问设备硬件能力.
 * 每个模块的功能参见对应的头文件.
 *
 * @param id 要获取的厂商接口模块 id.
 * @param module 要获取的模块填入此处.
 * @return 获取是否成功
 *   - WXERROR_NOT_FOUND: 找不到指定模块, 或者指定模块未实现
 *   - WXERROR_INVALID_ARGUMENT: module_out 为 NULL
 */
typedef wx_error_t (*wx_get_module_t)(const char* id,
                                      struct wx_module** module_out);

WX_END_DECLS
