#pragma once

#include "wmpf/error.h"
#include "wmpf/macros.h"

WX_BEGIN_DECLS

/**
 * @brief 异步操作
 *
 *
 */
typedef struct wx_operation* wx_operation_t;

typedef void (*wx_operation_callback_t)(wx_operation_t operation,
                                        void* user_data,
                                        wx_error_t result);

/**
 * @brief 异步等待操作
 * 该函数不会阻塞当前线程, 操作完成时将调用 callback.
 * 该函数在调用完成 callback 之后将会自动释放 wx_operation_t 对象,
 * 您不应该在调用该函数后调用 wx_operation_destroy.
 *
 * @param operation 要异步等待的操作
 * @param callback 操作完成时被调用的回调
 * @param user_data 回调用户自定义参数
 */
WX_API void wx_operation_await(wx_operation_t operation,
                               wx_operation_callback_t callback,
                               void* user_data);

/**
 * @brief 同步等待操作.
 * 该函数会阻塞当前线程直到操作完成.
 * 操作完成后 wx_operation_t 对象将被自动释放, 此时你不应该再调用
 * wx_operation_destroy 释放对象.
 * 如果操作等待超时, wx_operation 对象将不会自动释放, 如果你不再继续调用
 * wx_operation_wait 或 wx_operation_await 等待操作结束, 你需要手动调用
 * wx_operation_destroy 释放 wx_operation 对象.
 *
 * @param operation 要同步等待的操作.
 * @param timeout_ms 同步等待超时时间, 0 表示不设置超时时间.
 * @return 操作是否成功
 *   - WXERROR_TIMEOUT: 操作等待超时, 您可能需要调用 wx_operation_destroy
 * 函数释放 wx_operation 对象
 *   - 其他错误: 操作本身的返回值.
 */
WX_API wx_error_t wx_operation_wait(wx_operation_t operation,
                                    uint32_t timeout_ms);

/**
 * @brief 建议取消操作.
 *
 * 该函数调用后，操作可能会被取消.
 * 该函数会自动释放 wx_operation 对象, 即该函数调用后不能再调用
 * wx_operation_wait 或 wx_operation_await 或 wx_operation_destroy 方法.
 *
 * @param operation
 * @return WX_API
 */
WX_API wx_error_t wx_operation_cancel(wx_operation_t operation);

/**
 * @brief 释放 wx_operation 对象
 * 如果你不调用 wx_operation_await 或者在调用 wx_operation_wait 之后返回
 * WXERROR_TIMEOUT 时不再继续调用等待方法, 你应该调用 wx_operation_destroy 释放
 * wx_operation 对象.
 *
 * @param operation 要被释放的 wx_operation 对象.
 */
WX_API void wx_operation_destroy(wx_operation_t operation);

WX_END_DECLS
